#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/BoostWhenAnswers.pm,v 1.11 2005/06/10 16:17:16 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::Modules::BoostWhenAnswers>> - Aranea module for boosting
answers to `when' questions

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

This Aranea module serves as an answer booster specialized for `when'
questions.  The central idea is to incorporate rather simple
heuristics to filter out obviously wrong answers.

=cut

###############################################################################
#>
#< Header
package Aranea::Modules::BoostWhenAnswers;

use strict;
use warnings;

use IO::Handle;

use Aranea::XML::Aranea;

our $MONTHS;

#>

#< BEGIN
BEGIN {
    $MONTHS = '(January|February|March|April|May|June|July|August|September|October|November|December|Jan\.?|Feb\.?|Mar\.?|Apr\.?|May\.?|Jun\.?|Jul\.?|Aug\.?|Sept\.?|Oct\.?|Nov\.?|Dec\.?)';
}
#>

#< STDIN Handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    $Aranea->insert_pipeline_tag('$Id: BoostWhenAnswers.pm,v 1.11 2005/06/10 16:17:16 jimmylin Exp $ ');

    # do nothing if there are no entries
    return $Aranea if $Aranea->request_list()->entries_count() == -1;

    my $query = $Aranea->query()->text();
    # do nothing if it isn't a "when" question
    return $Aranea unless $query =~ /^When/i;

    # "When was..." questions cannot be answered without a year.  If
    # it occured in the past, then we must have a fully qualified
    # date.  Note that this isn't true for all "when" questions, e.g.,
    # "When is Boxing Day?"
    if ( $query =~ /When was/i ) {
      $Aranea->candidate_list()
	->filter_candidates(sub { shift !~ /^$MONTHS\s+\d+$/ });
    }

    # these recent years are so common that we should throw them out
    $Aranea->candidate_list()
	->filter_candidates(sub { shift !~ /(1999|2000|2001|2002)/ });

    # Single and double digit numbers cannot possibly be a valid date.
    # Even if it indicated a year, it should be qualified with 'AD',
    # 'BC', etc.
    $Aranea->candidate_list()->filter_candidates(sub { shift !~ /^\d\d?$/ });

    # In the same vein, if you have a two digit number, there should
    # be a month associated with it (or something like AD, BC, etc.)
    # e.g., q1197 What date did Neil Armstrong land on the moon?
    # 'Apollo 11'
    $Aranea->candidate_list()
	->filter_candidates(sub {
				my $s = shift;
				not ($s =~ / \d\d?,? / and
				     $s !~ /$MONTHS/ and
				     $s !~ /(ADE?|BCE?)/ ) });

    return $Aranea if $Aranea->candidate_list()->entries_count() == -1;

    my $entries = $Aranea->candidate_list()->entries();
    my $e = @$entries[0];
    my $s = $e->candidate();

    # A lone ordinal number probably indicates a century
    $e->candidate("$s century") if $s =~ /^\d+th$/;

    # If there is other trash along with a four digit number (year),
    # and it isn't a month, then kill it, e.g., q336 When was
    # Microsoft established? '1975 by Bill Gates'.
    if ( $s =~ /\d\d\d\d/ ) {
	my $year = $&;
	$e->candidate($year) unless $s =~ /$MONTHS/i;
    }

    # If the first candidate is in the form "Month \d\d", then we are
    # probably looking for a complete date, e.g., "Month Day, Year."
    # Assuming this is the case, look in the top five entries to see
    # if we can find the correct year.
    #
    # The known exception to this is a question like "When is
    # St. Patrick's Day?"  In this case, it is probably very unlikely
    # to find "March 17, 2002" in the top five answers... so we should
    # be safe.

    if ( $s =~ /$MONTHS\s+\d+/ ) {
	for my $i ( 1 .. 4 ) {
	    last if $i > $#{@$entries};
	    my $cur = @$entries[$i];
	    if ( $cur->candidate() =~ /$s, \d\d\d\d/ ) {
		my $new = $&;
		#print "==$new==\n";
		$e->candidate($new);
		last;
	    }
	}
    }

    return $Aranea;
}
#>

1;

