#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/BoostWhereAnswers.pm,v 1.12 2005/06/10 16:17:16 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::Modules::BoostWhereAnswers>> - Aranea module for boosting
answers to `where' questions

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

This Aranea module serves as a knowledge-based answer booster
specialized for `where' questions.  The idea is to incorporate a
named-entity extractor and employ other heuristics to filter out
obviously wrong answers.  The critical insight is to perform this of
filtering B<I<after>> the redundancy-based answer extraction process
has run its course, B<I<not>> during the process.  This way,
less-than-perfect named entity detection will not hurt system
performance.  In general, named-entity recall is more important than
precision because Web-redundancy serves to boost precision.

=cut

###############################################################################
#>
#< Header
package Aranea::Modules::BoostWhereAnswers;

use strict;
use warnings;

use IO::Handle;

use Aranea::XML::Aranea;
use Aranea::Patterns::Persons qw($BODY_PART);

our $STATES;
our $STATES_ABBREV;

#>

#< BEGIN
BEGIN {
    $STATES = '(?:Alabama|Alaska|Arizona|Arkansas|California|Colorado|Connecticut|Delaware|Florida|Georgia|Hawaii|Idaho|Illinois|Indiana|Iowa|Kansas|Kentucky|Louisiana|Maine|Maryland|Massachusetts|Michigan|Minnesota|Mississippi|Missouri|Montana|Nebraska|Nevada|New Hampshire|New Jersey|New Mexico|New York|North Carolina|North Dakota|Ohio|Oklahoma|Oregon|Pennsylvania|Rhode Island|South Carolina|South Dakota|Tennessee|Texas|Utah|Vermont|Virginia|Washington|West Virginia|Wisconsin|Wyoming)';

    $STATES_ABBREV = '(?:AL|AK|AS|AZ|AR|CA|CO|CT|DE|DC|FM|FL|GA|GU|HI|ID|IL|IN|IA|KS|KY|LA|ME|MH|MD|MA|MI|MN|MS|MO|MT|NE|NV|NH|NJ|NM|NY|NC|ND|MP|OH|OK|OR|PW|PA|PR|RI|SC|SD|TN|TX|UT|VT|VI|VA|WA|WV|WI|WY)';
}
#>

#< STDIN handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    $Aranea->insert_pipeline_tag('$Id: BoostWhereAnswers.pm,v 1.12 2005/06/10 16:17:16 jimmylin Exp $ ');

    # check to see if there is work to be done...
    return $Aranea if $Aranea->request_list()->entries_count() == -1;
    return $Aranea if $Aranea->candidate_list()->entries_count() == -1;

    my $query = $Aranea->query()->text();

    # do nothing if it isn't a "where" question
    return $Aranea unless $query =~ /^Where/i;

    my $entries = $Aranea->candidate_list()->entries();

    # sort by score descending
    @$entries = sort { $b->{score} <=> $a->{score} } @$entries;
    my $new_entries = [];

    # Sometimes in the top answer we'll get a NP PP combination.  If
    # the PP is headed by 'in', then we assume that it specficies the
    # location, and we'll throw out the NP.  For example, the top
    # answer to q1526 'Where are diamonds mined?' was 'rebels in
    # Sierra Leone.'  This procedure would fix this answer and make it
    # exact.

    my $e = @$entries[0];
    my $s = $e->candidate();

    if ( $s =~ /in (.*)/ ) {
	my $new = $1;
	$e->candidate($new);
    }

    # If body parts appear in the top five, then boost it up to the
    # top.  This handles questions like q393 "Where is your corpus
    # callosum?"

    my $top_score = $e->score();
    my $i = 0;
    for my $entry (@$entries) {
	$i++;
	my $candidate = $entry->candidate();
	my $score = $entry->score();

	if ( $candidate =~ /^$BODY_PART$/i and $i < 5 ) {
	    my $new_entry = Aranea::XML::CandidateEntry->new($&, $score+$top_score);
	    $new_entry->support($entry->support());
	    #print "##$candidate--$top_score\n";
	    push @$new_entries, $new_entry;
	} else {
	    #print "##!$candidate\n";
	    push @$new_entries, $entry;
	}
    }

    # Since we've gone through all the entries in the 'body part'
    # check, we can go ahead and reset the candidate list

    $Aranea->candidate_list()->entries($new_entries);
    $entries = $new_entries;
    $new_entries = [];

    # A heuristic that works very well is to look for patterns of the
    # form "city, state" in the top ten answers.  If such a pattern is
    # found, assume that it is the correct answer.  The state is easy
    # to detect because it is a closed class (both full form and
    # standard 2 letter abbreviations); any sequence of capitalized
    # tokens is assumed to be a city.  This heuristic works very well
    # because "city, state" is the perfect granularity for an answer
    # to a question about locations inside the United States.

    # This heuristic does not extend to locations outside the
    # US. Sometimes, the country alone is provides an answer with
    # sufficient granularity, both often times it does not.  Consider
    # q1326, "where are the British crown jewels kept."  I do not
    # believe that "England" is good enough as an answer (therefore,
    # filtering by country does not work all the time.)  However,
    # consider q698, "where is Ocho Rios."  I believe that "Jamaica"
    # suffices as an answer.
    $i = 0;
    for my $entry (@$entries) {
	$i++;
	last if $i > 20;
	my $candidate = $entry->candidate();
	my $score = $entry->score();

	if ( $candidate =~ /([A-Z][^ ]+ )*[A-Z][^,]+, $STATES/o ) {
	    # note, compile the regexp for speed
	    my $new_entry = Aranea::XML::CandidateEntry->new($&, 2*$score);
	    $new_entry->support($entry->support());

	    push @$new_entries, $new_entry;
	}

	# We've normalized the state abbreviations, so this isn't
	# necessary...
	if ( $candidate =~ /([A-Z][^, ]+ )*[A-Z][^,]+, $STATES_ABBREV/o ) {
	    # note, compile the regexp for speed
	    my $new_entry = Aranea::XML::CandidateEntry->new($&, 1*$score);
	    $new_entry->support($entry->support());

	    push @$new_entries, $new_entry;
	}
    }

    $Aranea->candidate_list()->entries($new_entries)
	if $#{@$new_entries} != -1;

    return $Aranea;
}
#>

1;
