#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/CleanupForTREC.pm,v 1.17 2005/06/10 16:17:16 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::Modules::CleanupForTREC>> - Aranea module for cleaning up
the candidate answers

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

This Aranea module performs general cleanup of the candidate answers,
specially in the context of 'exact answers' for TREC.  Occasionally,
an answer would give too much information, e.g., "Bell in 1876" as the
answer to "When was the telephone invented?"  This module performs
some sanity checks and eliminates the extra information.

=cut

###############################################################################
#>
#< Header
package Aranea::Modules::CleanupForTREC;

use strict;
use warnings;

use IO::Handle;

use Aranea::XML::Aranea;
use Aranea::Common qw(:default);

#>

#< STDIN handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    $Aranea->insert_pipeline_tag('$Id: CleanupForTREC.pm,v 1.17 2005/06/10 16:17:16 jimmylin Exp $ ');

    # Do nothing if there are no entries: no request means that an
    # exact answer was retrieved via database lookup; hence, no need
    # for cleanup
    return $Aranea if $Aranea->request_list()->entries_count() == -1;

    # Naturally, do nothing if there are no candidates.
    return $Aranea if $Aranea->candidate_list()->entries_count() == -1;

    my $q = $Aranea->query()->text();
    my $candidate_entries = $Aranea->candidate_list()->entries();
    @$candidate_entries = 
	sort { $b->score() <=> $a->score() } @$candidate_entries;

    my $first = @$candidate_entries[0];
    my $candidate = $first->candidate();

    # Clean up partial dates.  Sometimes we get a fragment of the day
    # and year without the month, e.g., q445 When was the last major
    # eruption of Mount St. Helens? '18, 1980' In a case like this, we
    # want remove the actual day and only keep the year.

    $candidate = $1 if $candidate =~ /^\d+, (\d+)$/;

    # Make sure that the answer doesn't have any extra verbs or
    # adverbs, e.g., in q94 'Who was the oldest U.S. president?'  one
    # of the answers extracted was 'Ronald Reagan turned'.  After POS
    # tagging, we can chop off trailing (and leading) verbs/adverbs.

    # NOTE: POS tagging messes up spacing with numbers, so ignore if
    # there are numbers

    # NOTE: took out POS tagging because it isn't accurate enough to
    # use.  They require context words to get an accurate tag ---
    # context which isn't available in exact answers (July 21, 2003)

    #if ( $candidate !~ /\d/ and $candidate =~ / / ) {
	# POS tag and chop off verbs/adverbs
	#$candidate = pos_tag($candidate);
	#while ($candidate =~ s/ [a-z][^\/]+\/(RB|VB.)\s*$// ) {};
	#while ($candidate =~ s/^[a-z][^\/]+\/(RB|VB.)\s*// ) {};
	# be conservative and don't chop of capitalized words

	# now remove all tags and restore candidate
	#$candidate = strip_pos_tags($candidate);
    #}

    # Another issue is the 'exactness' of 'when' questions, e.g.,
    # q1349 When was the telephone invented?  a likely answer is
    # something like 'Graham Bell in 1876.'  If we a 'in \d\d\d\d' in
    # the context of a 'when' question, assume that the year is the
    # exact answer.

    $candidate = $1 if $q =~ /^When/i and $candidate =~ /in (\d\d\d\d)/;

    # restore the modified candidate
    $first->candidate($candidate);

    return $Aranea;
}
#>

1;
