#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/Collect.pm,v 1.14 2005/06/10 16:17:16 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::Modules::Collect>> - Aranea module for collecting
candidates that are substrings of each other

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

This Aranea module removes answer candidates that are substrings of
other answer candidate.  Comments in the source code discuss in detail
heuristics that tend to work and other caveats.

=cut

###############################################################################
#>
#< Header
package Aranea::Modules::Collect;

use strict;
use warnings;

use IO::Handle;

use Aranea::XML::Aranea;

#>

#< STDIN handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    $Aranea->insert_pipeline_tag('$Id: Collect.pm,v 1.14 2005/06/10 16:17:16 jimmylin Exp $ ');

    # if there are no requests, then there is nothing to do
    return $Aranea if $Aranea->request_list()->entries_count() == -1;

    my $entries = $Aranea->candidate_list()->entries();

    # don't apply this procedure on too many candidates, for speed
    # reasons. default is top 30 candidates
    my $topn = $#{@$entries} > 30 ? 30 : $#{@$entries};
    my $end_idx = $#{@$entries};

    # Feb. 7, 2003 -- I discovered that this module is very very
    # strangly written.  The entries are sorted by score asc, and all
    # operations loop over the entries from the end of the array.  Why
    # not sort desc and loop over the front of the array?
    @$entries = sort { $a->score() <=> $b->score() } @$entries;

    # If any lower ranking entry is the substring of any higher
    # ranking entry, then remove it.  We can safely do this because we
    # assume that a longer answer with a higher score is always better
    # than a shorter answer with a lower score.
    for (my $i = $end_idx; $i > $end_idx-$topn; $i-- ) {
	my $e1 = @$entries[$i];
	next unless $e1;                # entry might have already been removed
	my $c1 = $e1->candidate();

	# inner loop: check all subsequent candidates
	for (my $j = $i-1; $j > $end_idx-$topn; $j-- ) {
	    my $e2 = @$entries[$j];
	    next unless $e2;
	    my $c2 = $e2->candidate();

	    @$entries[$j] = undef if $c1 and $c2 and $c1 =~ /\Q$c2\E/i;
	    # if the lower candidate is a substring of the higher
	    # candidate, then remove the lower candidate: just undef
	    # the array entry.  yes, `the right way' to do this is to
	    # splice out the entry, but this will suffice.
	}
    }

    # Take the case:
    #   1. A
    #   2. A B
    #
    # We can safely discard (1) and assign its score to (2).  However,
    # this can only be done if the candidates are adjacent.

    # August 7, 2002 -- Technically, this is not absolutely true.
    # Ocassionally, the lower answer will be less exact than the
    # higher one, e.g., '1984' vs. '1984 the author'.  Killing the
    # higher/better one will actually reduce the quality of the
    # answer.  The solution is to leave this heuristic as is and count
    # on answer cleaners and answer boosters to solve this problem.

    # On the other hande, consider the case:
    #   1. A
    #   2. C
    #   3. A B
    #
    # We can't be sure what to do... If we kill A, then we might be
    # dropping a totally valid answer.  If we bring (3) up to (1),
    # that has the possibility of introducing lots of noise... what if
    # instead of one intervening candidate, there were two, or three,
    # or even ten? In other words, we want this procedure to preserve
    # relative ranking.

    $end_idx = $#{@$entries};
 OUTER: for (my $i = $end_idx; $i > $end_idx-$topn; $i-- ) {
	my $e1 = @$entries[$i];
	next unless $e1;
	my $c1 = $e1->candidate();

	my $j = $i-1;
	# skip empty entries (removed by the previous procedure)
	$j-- while not @$entries[$j];

	my $e2 = @$entries[$j];
	next unless $e2;
	my $c2 = $e2->candidate();

	if ( $c1 and $c2 and $c2 =~ /\Q$c1\E/i ) {
	    $e2->score($e1->score());  # exchange scores
	    @$entries[$i] = undef;
	    next OUTER;
	}
    }

    # Now collect up the remaining candidates...  Basically, we keep
    # only the top n candidates.  Note that the candidates are sorted in
    # ascending order according to score, so we have to take the
    # candidates from the rear of the list.

    my $new_entries = [];
    for (my $i = $end_idx; $i >= $end_idx-$topn; $i-- ) {
	my $entry = @$entries[$i];

	next unless $entry and $entry->candidate();
	push @$new_entries, $entry;
    }

    $Aranea->candidate_list()->entries($new_entries);

    return $Aranea;
}
#>

1;
