#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/FilterClosedClass.pm,v 1.7 2005/06/10 16:17:16 jimmylin Exp $
#
#>
#< POD

###############################################################################

=head1 NAME

B<C<Aranea::Modules::BoostClosedClassAnswers>> - Aranea module for
boosting closed class answers.

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

This Aranea module ensures that closed class answers are valid,
throwing away any invalid candidates.  A closed classed answer is one
whose form can be either exhaustively enumerated or easily detected.
For example, the answer to a "What state..." question must be a state,
and since there are only 50 states, the module can check to see if the
answer candidate is a valid state.  "What president..." and "What
country..." questions are two more classes than fall under this
category.  Another type of closed class answer involves extracting the
question focus, e.g., the answer to a "How many feet..." question must
have the unit `feet', so the module can verify correct units and throw
out all invalid forms.

=cut

###############################################################################

#>
#< Header

package Aranea::Modules::FilterClosedClass;

use strict;
use warnings;

use Data::Dumper;
use IO::Handle;

use Aranea::XML::Aranea;
use Aranea::Patterns::Dates qw($MONTH);
use Aranea::Patterns::Locations qw($COUNTRY $STATE $CANADIAN_PROVINCE $POSTAL_CODE $COLLEGE $ADJ_TO_LOCATION_keys);
use Aranea::Patterns::Misc qw($LANGUAGE);
use Aranea::Patterns::Persons qw($US_PRESIDENT);

our %WORLD_CITIES;

#>

#< BEGIN
BEGIN {
    dbmopen(%WORLD_CITIES, "pm/Aranea/Modules/world-weather-cities.dbm", 0666)
	or die "can not open DBM file\n";
}
#>

#< STDIN Handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    # July 31, 2003 -- split 'FilterClosedClass' from
    # 'BoostClosedClassAnswers'; see the latter module for detailed
    # explanation

    $Aranea->insert_pipeline_tag('$Id: FilterClosedClass.pm,v 1.7 2005/06/10 16:17:16 jimmylin Exp $ ');

    # even if there are no requests, there may be work to do,
    # e.g., q387 "What year did Montana become a state?":
    # the answer is retrieved by database lookup via 50states.com,
    # but we want to extract only the year.

    my $query = $Aranea->query()->text();
    my $tagged = $Aranea->query()->tagged();
    my $entries = $Aranea->candidate_list()->entries();

    # filter zip codes; they take a distinctive 5-digit form
    $Aranea->candidate_list()->filter_candidates(\&match_zip_code)
	if $query =~ /^What is the zip code /i or
	   $query =~ /^What zip code /i;

    # filter against a list of cities (from Weather.com)
    $Aranea->candidate_list()->filter_candidates(\&match_city)
	if $tagged =~ /^(What|Which)\/[^ ]+ ([^\/]+\/(JJ|NNP) )*city/i or
	   $query =~ /^What is the ([^ ]+ )?city/i;

    # Handle questions like the following:
    #     q1514 "What is Canada's most populous city?"
    #     q1779 "What is Australia's oldest city?"
    $Aranea->candidate_list()->filter_candidates(\&match_city)
	if $tagged =~ qr{^(What|Which)/[^\s]+\sis/VBZ\s
			 ([^/]+/(NNS?|NNPS?|RB[RS]?|JJ[RS]?)\s)+city/NN}xi;

    # filter against a list of colors
    $Aranea->candidate_list()->filter_candidates(\&match_color)
	if $query =~ /^(What|Which) ([^ ]+ )?colors?/i or
	   $query =~ /^(What|Which) are the colors?/i;

    # filter against a list of colleges/universities
    $Aranea->candidate_list()->filter_candidates(\&match_college)
	if $query =~ /^(What|Which) (college|university)/i;

    # filter against a list of continents
    $Aranea->candidate_list()->filter_candidates(\&match_continent)
	if $query =~ /^(What|Which) continent/i;

    # Filter against a list of countries (from Infoplease/World
    # Factbook).  NOTE: we have to handle "What countries" questions
    # separately...  We need an exactify procedure that matches a list
    # of countries, not just a single one
    $Aranea->candidate_list()->filter_candidates(\&match_country)
	if $tagged =~ /^(What|Which)\/[^ ]+ ([^\/]+\/(JJ|NNP) )*(country|nation)\//i;
    # beware of things like q1499 "Which African country's major
    # export is coffee?"

    # Handle questions like:
    #     q1615 "What is Africa's largest country?"
    $Aranea->candidate_list()->filter_candidates(\&match_country)
	if $tagged =~ qr{^(What|Which)/[^\s]+\sis/VBZ\s
			 ([^/]+/(NNS?|NNPS?|RB[RS]?|JJ[RS]?)\s)+(country|nation)/NN}xi;

    # filter against the four hemisphers
    $Aranea->candidate_list()->filter_candidates(\&match_hemisphere)
	if $query =~ /^(What|Which) hemisphere/i;

    # filter against a list of languages (from CIA World Factbook)
    $Aranea->candidate_list()->filter_candidates(\&match_language)
	if $query =~ /^(What|Which) language/i;

    # Filter against a list of months
    $Aranea->candidate_list()->filter_candidates(\&match_month)
	if $query =~ /^(What|Which) month /i;

    # filter against a list of nationality
    $Aranea->candidate_list()->filter_candidates(\&match_nationality)
	if $query =~ /^(What|Which) nationality/i;

    # filter against a list of oceans
    $Aranea->candidate_list()->filter_candidates(\&match_ocean)
	if $query =~ /^(What|Which) ocean/i;

    # filter against a list of planets in the Solar System
    $Aranea->candidate_list()->filter_candidates(\&match_planet)
	if $query =~ /^(What|Which) planet/i;

    # filter against a list of U.S. presidents
    $Aranea->candidate_list()->filter_candidates(\&match_president)
	if ( $query =~ /^Who/i and $query =~ /U\.?S\.? president/i ) or
	    $query =~ /^(What|Which) (U\.?S\.?(A\.?)? )?president/i;

    # filter against a list of provinces (Canadian)
    $Aranea->candidate_list()->filter_candidates(\&match_province)
	if $query =~ /^(What|Which) (Canadian )?province /i;

    # filter against a list of states.
    $Aranea->candidate_list()->filter_candidates(\&match_state)
	if $query =~ /^(What|Which) (U\.S\. )?state/i or
           $query =~ /^(In )?(what|which) (U\.S\. )?state/i;

    # filter against a list of sports (this one might be
    # overtraining...)
    $Aranea->candidate_list()->filter_candidates(\&match_sport)
	if $query =~ /sport/i;

    # filter against years. NOTE that we use the filter_entries
    # method, not the filter_candidates method.
    $Aranea->candidate_list()->filter_entries(\&match_year)
	if $query =~ /^(What|Which) year/i;

    #print Dumper($Aranea->candidate_list()->entries());

    return $Aranea;
}
#>

#< match_city
sub match_city {
    my $s = shift;

    $s = $1 if $s =~ /(.*?) City$/;
    $s = $1 if $s =~ /(.*?), ($STATE|$POSTAL_CODE)$/;

    $WORLD_CITIES{$s} and $s !~ /^($COUNTRY|$STATE|$CANADIAN_PROVINCE)$/;
}
#>
#< match_contintent
sub match_continent {
    shift =~ /^(Asia|Africa|North\sAmerica|South\sAmerica|Australia|Europe|Antarctica)$/;
}
#>
#< match_college
sub match_college {
    shift =~ /^$COLLEGE$/;
}
#>
#< match_color
sub match_color {
    shift =~ /(^| )(red|orange|yellow|green|blue|indigo|purple|black|white|brown|gray|magenta|cyan|gold|turquoise)( |$)/i;
}
#>
#< match_country
sub match_country {
    shift =~ /^$COUNTRY$/;
}
#>
#< match_hemisphere
sub match_hemisphere {
    shift =~ /^(North(ern)?|Sourth(ern)?|East(ern)?|West(ern)?)$/;
}
#>
#< match_language
sub match_language {
    shift =~ /^$LANGUAGE$/;
}
#>
#< match_month
sub match_month {
    shift =~ /^$MONTH$/;
}
#>
#< match_nationality
sub match_nationality {
    shift =~ /^$ADJ_TO_LOCATION_keys$/;
}
#>
#< match_ocean
sub match_ocean {
    shift =~ /^(Atlantic|Pacific|Indian|Arctic|Southern)$/;
}
#>
#< match_planet
sub match_planet {
    shift =~ /^(Mercury|Venus|Earth|Mars|Jupiter|Saturn|Uranus|Neptune|Pluto)$/;
}
#>
#< match_president
sub match_president {
    shift =~ /(^$US_PRESIDENT$|^$US_PRESIDENT, $US_PRESIDENT$)/;

    # NOTE: POTUS gets back 'Calvin Coolidge, Herbert Clark Hoover'
    # for the answer to q1678 "Who was the U.S. president in 1929?"
    # We don't want to accidentally throw this out...
}
#>
#< match_province
sub match_province {
    shift =~ /^$CANADIAN_PROVINCE$/;
}
#>
#< match_sport
sub match_sport {
    shift =~ /^(baseball|basketball|football|soccer|hockey)$/;
}
#>
#< match_state
sub match_state {
    shift =~ /^$STATE$/;
}
#>
#< match_year
# Note: to be used with 'filter_entries'
sub match_year {
    my $entry = shift;
    my $candidate = $entry->candidate();

    return unless $candidate =~ /\d\d\d\d/;
    my $year = $&;

    return if $year == 1999 or $year == 2000 or
	$year == 2001 or $year == 2002;

    $entry->candidate($year);

    return 1;
}
#>
#< match_zip_code
sub match_zip_code {
    shift =~ /^\d\d\d\d\d$/;
}
#>

#< END
END {
    dbmclose %WORLD_CITIES or die "Couldn't open DBM file!\n";
}
#>

1;
