#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/FilterSimple.pm,v 1.7 2005/06/10 16:17:16 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::Modules::FilterSimple>> - Aranea module for filtering answer
candidates

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

This Aranea module performs filtering of the answer candidates.
Candidates that either begin or end with stopwords are thrown away
(with the exception of candidates beginning with `New', as in `New
York'.)  Candidates that contain words present in the question are
also thrown out (with the exception of the question focus, e.g., in a
question like `How many turtles...', having `turtles' in the answer is
okay.

=cut

###############################################################################
#>
#< Header
package Aranea::Modules::FilterSimple;

use strict;
use warnings;
use lib qw(lib pm);                         # needed for command-line mode

use IO::Handle;

use Aranea::XML::Aranea;
use Aranea::XML::CandidateEntry;
use Aranea::Common qw(:default);
use Aranea::Lexicon::Lexicon qw(:default);

our %ABBREV;

#>

#< BEGIN
BEGIN {
    # a hash of tokens that can legitimately be trailed by a period
    # without breaking the sentence.

    %ABBREV = (
	       "Alex." => 1,
	       "Adm." => 1,
	       "Brig." => 1,
	       "Capt." => 1,
	       "Cie." => 1,
	       "Cmdr." => 1,
	       "Col." => 1,
	       "Cpl." => 1,
	       "Dr." => 1,
	       "Drs" => 1,
	       "Fr." => 1,
	       "Ft." => 1,
	       "Gen." => 1,
	       "Gov." => 1,
	       "Lt." => 1,
	       "Maj." => 1,
	       "Messrs." => 1,
	       "Mme." => 1,
	       "Mr." => 1,
	       "Mrs" => 1,
	       "Ms." => 1,
	       "Mt." => 1,
	       "Ph." => 1,
	       "Prof." => 1,
	       "Prop." => 1,
	       "Pte." => 1,
	       "Pvt." => 1,
	       "Rep." => 1,
	       "Reps." => 1,
	       "Rev." => 1,
	       "Sen." => 1,
	       "Sens." => 1,
	       "Sgt." => 1,
	       "St." => 1,
	       "Ste." => 1,
	       "Jan." => 1,
	       "Feb." => 1,
	       "Febr." => 1,
	       "Mar." => 1,
	       "Apr." => 1,
	       "Jun." => 1,
	       "Jul." => 1,
	       "Aug." => 1,
	       "Sep." => 1,
	       "Sept." => 1,
	       "Oct." => 1,
	       "Nov." => 1,
	       "Dec." => 1,
	      );
}
#>

#< STDIN Handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    $Aranea->insert_pipeline_tag('$Id: FilterSimple.pm,v 1.7 2005/06/10 16:17:16 jimmylin Exp $ ');

    # if there are no requests, then there is nothing to do
    return $Aranea if $Aranea->request_list()->entries_count() == -1;

    my $query = $Aranea->query()->text();

    my $entries = $Aranea->candidate_list()->entries();
    my $new_entries = [];

    # In 'how many' questions, the question focus often appears in the
    # answer.  Therefore, we do not want to filter out candidates that
    # contain the focus,
    #   e.g., q1391 How many feet in a mile?  '5,280 feet'

    if ( $query =~ /^How many/i ) {
	my $qfocus = $Aranea->query()->qfocus();
	$query =~ s/$qfocus//g if $qfocus;
    }

    my @query_words = split(/-|\/|\s+/, $query);
    # remember to split things like 'Winnie-the-Pooh' and 'Mason/Dixon'

    for my $word ( @query_words ) {
	$word =~ s/\'[Ss]$//g;          # strip possessive ending
	$word =~ s/\W+$//g;             # kill puncts
	$word =~ s/^\W+//g;
	# normalize abbreviations so that they don't have the dots
	$word =~ s/\.//g if $word =~ /[A-Z]\.[A-Z]/;

	my $stemmed = stem($word);
	# we don't know POS, so just take the first one.
	$word = $stemmed->[0][0] if $stemmed->[0][0];

	$word = derivational_stem($word);
    }

    # common variants of words that shouldn't appear in the answer if
    # they appear in the question.
    push @query_words, '1st' if $query =~ /first/;
    push @query_words, 'goddess' if $query =~ /god/;
    push @query_words, 'god' if $query =~ /goddess/;
    push @query_words, $1 if $query =~ /(\d\d\d\d)('s|s)/;

    if ( $query =~ /U\.?S\.?(A\.?)?/ ) {
	push @query_words, 'United';
	push @query_words, 'States';
    }

 GRAM: foreach my $entry (@$entries) {
	my $gram = $entry->candidate();

	# candidates shouldn't have '>' in them; it usually indicates
	# a dmoz directory listing.
	next if $gram =~ / > /;

	# stopwords filtering: if the candidate consists entirely of
	# stopwords, if it begins or ends with a stopword, skip this
	# candidate by going to the next iteration of the GRAM loop

	# removing candidates beginning with stopwords was a decision
	# made with plenty of deliberation.  The original idea was
	# that prepositional phrases could be valid answers, e.g.,
	# "Where is Belize located?" -- "in Central America."
	# However, upon further consideration, I presumed that judges
	# will accept "Central America" as well for an "exact answer."
	# Therefore, it's better to be more conservative... keeping
	# the stopword might lead to problems later on, e.g., tiling
	# and substring collection.

	my @swords = split(/\s+/, $gram);
	# throw away candidate if it ends in a stopword
	next GRAM if is_stopword(pop(@swords));

	# throw away candidate if it begins in a stopword.  one
	# caveat, though: don't through away something that begins
	# in 'New', because we don't want to toss out 'New York,'
	# 'New Jersey', etc.
	my $front = shift(@swords);
	next GRAM if is_stopword($front) and $front ne 'New';

	# throw away candidate if it consists entirely of stopwords
  	my $all_stopwords = 1;
	foreach my $word (@swords) {
	    $all_stopwords = 0, last if !is_stopword($gram)
	}
	#my $all_stopwords_t = grep { is_stopword($_) } @swords;
	#print "$all_stopwords, $all_stopwords_t, $#{@swords}\n";
	next GRAM if $all_stopwords && (scalar(@swords) > 0);

	# In the following block, we want to throw away all candidates
	# that contain words from the question.  The notable
	# exception, however, involves question focus (see comment
	# above).

	my @words = split(/-|\/|\s+/, $gram);
	# NOTE: we want to split on dashes and slashes also.
	# Otherwise, the answer to something like "What kind of animal
	# was Winnie the Pooh?" (q408) would be "Winnie-the-Pooh"

	foreach my $word ( @words ) {

	    # NOTE: we have to do the exact same mutations to the
	    # candidates as we did to the query tokens.

	    $word =~ s/\'[Ss]$//g;             # strip possessive ending
	    $word =~ s/\W+$//g;                # kill puncts
	    $word =~ s/^\W+//g;
	    $word =~ s/\.//g if $word =~ /[A-Z]\.[A-Z]\./;

	    my $match_count;

	    # check to see if the candidate word is one of the query terms
	    $match_count = grep(/^\Q$word\E$/i, @query_words);
	    next GRAM if $match_count != 0;

	    my $stemmed = stem($word);
	    # we don't know POS, so just take the first one.
	    $word = $stemmed->[0][0] if $stemmed->[0][0];

	    # do the same thing, now with stemmed verison.
	    $match_count = grep(/^\Q$word\E$/i, @query_words);
	    next GRAM if $match_count != 0;

	    $word = derivational_stem($word);
	    # try again after horrible derivational stemming
	    $match_count = grep(/^\Q$word\E$/i, @query_words);
	    next GRAM if $match_count != 0;

	    # NOTE: this loop may look inefficient, but for the wrong
	    # reason!  All the grep'ing is actually pretty fast; the
	    # running time is dominated by the entire outer loop,
	    # i.e., having to iterate through individual tokens.
	}

	# candidates should not have internal periods, unless its a
	# valid abbreviation.
	if ( $gram =~ /(\w+\.) / ) {
	    next GRAM unless $ABBREV{$1};
	}

	# create a new entry for candidate
	push @$new_entries,
             Aranea::XML::CandidateEntry->new($gram, $entry->score());
    }

    $Aranea->candidate_list->entries($new_entries);

    return $Aranea;
}
#>

#< derivational_stem (this is *really* ugly)
sub derivational_stem {
    my $word = shift;

    # this is a very poor excuse for a derivational stemmer...
    # e.g., in q250, kill 'Mayan'
    #       in q272, kill 'Aboriginal'
    $word =~ s/([Cc])anadian$/$1anada/g;       # e.g., Canadian
    $word =~ s/an$/a/g;
    $word =~ s/al$/e/g;

    $word =~ s/ism$//;         # e.g., q237 'Hinduism'

    return $word;
}
#>

1;
