#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/FilterType.pm,v 1.10 2005/06/10 16:17:16 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::Modules::FilterType>> - Aranea module for performing
some simplistic typer filtering on candidate answers.

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

This Aranea module performs simplistic type filtering on the answer
candidates.  For example, for questions that should have numerical
answers, any candidate without a numeral is thrown away; other
questions require answers that I<cannot> contain numerical answers.
For questions that require proper nouns, any candidate without a
capitalized letter is thrown away.

=cut

###############################################################################
#>
#< Header
package Aranea::Modules::FilterType;

use strict;
use warnings;

use IO::Handle;

use Aranea::Patterns::Dates qw($DATE);
use Aranea::XML::Aranea;

#>

#< STDIN Handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    $Aranea->insert_pipeline_tag('$Id: FilterType.pm,v 1.10 2005/06/10 16:17:16 jimmylin Exp $ ');

    # if there are no requests, then there is nothing to do
    return $Aranea if $Aranea->request_list()->entries_count() == -1;

    my $query = $Aranea->query()->text();
    my $tagq = $Aranea->query()->tagged();

    # with "what is the abbreivation" questions, the candidate much be
    # all caps, e.g., q446, q540, q1172
    $Aranea->candidate_list()->filter_candidates(\&keep_only_with_all_caps)
        if $query=~ /^What\s+is\s+the\s+abbreviation/i;

    # a variety of trigger terms for numerical answers: this is an
    # ad-hoc list derived from TREC9/10 questions.
    $Aranea->candidate_list()->filter_candidates(\&keep_only_with_digits)
        if $query=~ /^What\s+is\s+the\s+
                       (population|
                        (average\s+)?life\s+expectancy|
                        (average\s+)?life\s+span|
                        (freezing|melting|boiling)\s+point|
                        federal\s+minimum\s+wage|
                        size|
                        wingspan|
                        length|
                        salary|
                        zip\s+code|
                        telephone|
                        (average\s+)?depth|
                        (average\s+)?weight|
                        (average\s+)?speed|
                        (average\s+)?temperature|
                        sales\s+tax|
                        .*?rate|
                        percentage|
                        atomic\s+(weight|number)
                       )/xi;

    # for questions that involve time, require that the candidate
    # answers have some numerical component
    $Aranea->candidate_list()->filter_candidates(\&keep_only_with_digits)
        if $query=~ /^When /i or
	   $query =~ /^What year/i or
	   $query =~ /^What date/i;


    # Throw away all candidates that look like years for everything
    # except "when" questions
    $Aranea->candidate_list()->filter_candidates(\&toss_year)
        unless $query=~ /^(When |What year )/i;

    # for "how ADJ" type questions, e.g., "how fast," "how tall," "how
    # much," etc., require that the candidate answers have some
    # numerical component
    $Aranea->candidate_list()->filter_candidates(\&keep_only_with_digits)
	if $tagq =~ /^How\/WRB [^\/]+\/(RB|JJ)/i;

    # "where" and "who" questions should not have numerals in the answer
    $Aranea->candidate_list()->filter_candidates(\&throw_away_with_digits)
	if $query=~ /^Where / or $query=~ /^Who /;

    # Generally, "where" questions should all be answered with proper
    # names.  Notable exceptions:
    #
    # q393, Where is your corpus callosum?
    # q227, Where does dew come from?
    # q356, Where does chocolate come from?
    #
    # However, since the web is so big, we usually find capitalized
    # version of the answers, e.g., "Brain"
    #

    # I also checked all the "who" questions from TREC9/10, and
    # couldn't find any that can be answered by non-proper nouns.

    $Aranea->candidate_list()->filter_candidates(\&keep_only_with_caps)
	if $query =~ /^Where /i or $query =~ /^Who /i;

    # Kill leading and trailing lowercased words.  Takes care of
    # inexact answers like "James Earl Ray shot" and "called Poseidon",
    # where verbs are often folded into the proper name

    $Aranea->candidate_list()
	->filter_candidates(\&toss_proper_with_leading_lowercased_words)
	    if $query =~ /^Who /i or $query =~ /^Where /i;

    $Aranea->candidate_list()
	->filter_candidates(\&toss_proper_with_trailing_lowercased_words)
	    if $query =~ /^Who /i or $query =~ /^Where /i;

    # Keep only candidates that look like dates for when questions.

    $Aranea->candidate_list()
	->filter_candidates(\&keep_only_date)
	    if $query =~ /^When /i;

    # Double the score of candidates that have both numbers and the
    # question focus with "How many X" type questions.  e.g., q1391
    # How many feet in a mile? the candidate '5280 feet' would have
    # its score doubled.

    my $qfocus = $Aranea->query()->qfocus();
    if ( $query =~ /^How many /i and $qfocus ) {
	# the filtering criteria
	my $booster =
	    sub {
		my $entry = shift;
		my $candidate = $entry->candidate();

		# must have both the question focus and some digits
		$entry->score($entry->score() * 2) 
		    if $candidate =~ / $qfocus$/i and $candidate =~ /\d/;

		# we want to keep all entries, so always return true.
		return 1;
	    };
	$Aranea->candidate_list()->filter_entries($booster);
    }

    return $Aranea;
}
#>

#< various criteria for filtering
sub keep_only_with_digits { shift =~ /\d/ }
sub throw_away_with_digits { shift !~ /\d/ }
sub keep_only_with_caps { shift =~ /[A-Z]/ }
sub keep_only_with_all_caps { shift =~ /^[A-Z]+$/ }

sub toss_proper_with_leading_lowercased_words {
    shift !~ /^([a-z,'\-]+ )+[A-Z]/;
}
sub toss_proper_with_trailing_lowercased_words {
    shift !~ /[A-Z][a-z,'\-]+( [a-z]+)+$/;
}

sub toss_year {
    shift !~ /^[12]\d\d\d$/;
}

sub keep_only_date {
    my $s = shift;
    $s =~ /$DATE/ and $s !~ /200[234]/;
}
#>

1;
