#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/LookupGeneral.pm,v 1.15 2005/06/10 16:17:16 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::Modules::LookupGeneral>> - Aranea module for answering
location questions using database techniques

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

This module answers the following types of questions:

=over 4

=item *

The expansion of an acronym.

=item *

The area, capital, population, and currency of a country. (from
Infoplease, originally from World Factbook)

=item *

The tree, bird, flower, nickname, motto, population, capital, and area
of a state. (from 50States)

=item *

The symbol, discoverer, discovery date, atomic number, atomic weight,
melting point, and boiling point of an element. (from WebElements)

=back

=cut

###############################################################################
#>
#< Header
package Aranea::Modules::LookupGeneral;

use strict;
use warnings;

use Aranea::XML::Aranea;
use Aranea::XML::CandidateEntry;
use Aranea::Patterns::Locations qw($STATE $ADJ_TO_LOCATION_keys %ADJ_TO_LOCATION);
use Aranea::Resources::Gateway;
use Aranea::Resources::Acronyms;
use Aranea::Resources::50states;
use Aranea::Resources::Infoplease;
use Aranea::Resources::WebElements;

our %BIRTHSTONES2MONTH;
our %MONTH2BIRTHSTONES;
our $MONTHS;
our $BIRTHSTONES;

our %HEIGHT_HASH;
our $HEIGHT_OBJECTS;

#>

#< BEGIN
BEGIN {
    %BIRTHSTONES2MONTH =
	( 'garnet' => 'January',
          'rose Quartz' => 'January',
          'amethyst' => 'February',
          'onyx' => 'February',
	  'aquamarine' => 'March',
	  'bloodstone' => 'March',
	  'diamond' => 'April',
	  'quartz' => 'April',
	  'emerald' => 'May',
	  'chrysoprase' => 'May',
	  'alexandrite' => 'June',
	  'moonstone' => 'June',
	  'pearl' => 'June',
	  'ruby' => 'July',
	  'carnelion' => 'July',
	  'peridot' => 'August',
	  'sardonyx' => 'August',
	  'sapphire' => 'September',
	  'lapis' => 'September',
	  'opal' => 'October',
	  'tourmaline' => 'October',
	  'topaz' => 'November',
	  'citrine' => 'November',
	  'zircon' => 'December',
	  'turquoise' => 'December',
	);

    %MONTH2BIRTHSTONES =
	( 'January' => 'Garnet or Rose Quartz',
          'February' => 'Amethyst or Onyx',
	  'March' => 'Aquamarine or Bloodstone',
	  'April' => 'Diamond or Quartz',
	  'May' => 'Emerald or Chrysoprase',
	  'June' => 'Alexandrite or Moonstone or Pearl',
	  'July' => 'Ruby or Carnelion',
	  'August' => 'Peridot or Sardonyx',
	  'September' => 'Sapphire or Lapis',
	  'October' => 'Opal or Tourmaline',
	  'November' => 'Topaz or Citrine',
	  'December' => 'Zircon or Turquoise',
	);

    $MONTHS = '(January|February|March|April|May|June|July|August|September|October|November|December)';

    $BIRTHSTONES = '(Garnet|Rose Quartz|Amethyst|Onyx|Aquamarine|Bloodstone|Diamond|Quartz|Emerald|Chrysoprase|Alexandrite|Moonstone|Pearl|Ruby|Carnelion|Peridot|Sardonyx|Sapphire|Lapis|Opal|Tourmaline|Topaz|Citrine|Zircon|Turquoise)';

    %HEIGHT_HASH =
	( 'everest' => '29,035 ft. (8,850m)',
          'aconcagua' => '22,834 ft. (6,960m)',
          'mckinley' => '20,320 ft. (6,194m)',
          'denali' => '20,320 ft. (6,194m)',
          'kilimanjaro' => '19,340 ft. (5,995m)',
          'elbrus' => '18,510 ft. (5,642m)',
          'vinson massif' => '16,066 ft, (4,897)',
          'kosciusko' => '7,310 ft. (2,228)',
	  'chrysler building' => '1,046 ft. (319m)',
	  'petronas towers'  => '1,483 ft. (452m)',
	  'sears tower' => '1,450 ft. (442m)',
	  'empire state building' => '1,259 ft. (381m)',
	  'statue of liberty' => '152 ft.',
	  'gateway arch' => '630 ft.',
	  'space needle' => '605 ft.',
	  'eiffel tower' => '984 ft.',
        );

    $HEIGHT_OBJECTS = '(Everest|Aconcagua|McKinley|Denali|Kilimanjaro|Elbrus|Vinson|Kosciusko|Chrysler\s+Building|Petronas\s+(Towers?|Building)|Sears\s+(Tower|Building)|Empire\s+State\s+Building|Statue\s+of\s+Liberty|Gateway\s+Arch|Space\s+Needle|Eiffel\s+Tower)';

}
#>

#< STDIN handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    $Aranea->insert_pipeline_tag('$Id: LookupGeneral.pm,v 1.15 2005/06/10 16:17:16 jimmylin Exp $ ');

    my $query = $Aranea->query()->text();

    # heights of things
    if ( $query =~ /(^How tall|height of)/i and 
	 $query =~ /$HEIGHT_OBJECTS/i ) {
	my $obj = $&;
	$obj = 'petronas towers' if $obj =~ /Petronas\s+(Towers?|Building)/i;
	$obj = 'sears tower' if $obj =~ /Sears\s+(Tower|Building)/i;

	insert_answers($Aranea, "internal", [$HEIGHT_HASH{lc($obj)}])
    }

    # birthstones
    if ( $query =~ /^What /i and $query =~ /birthstone/i ) {
	insert_answers($Aranea, "internal", [$MONTH2BIRTHSTONES{$&}])
	    if $query =~ /$MONTHS/i;
	insert_answers($Aranea, "internal", [$BIRTHSTONES2MONTH{lc($&)}])
	    if $query =~ /$BIRTHSTONES/i;
    }

    if ( $query =~ /speed of light/ or
	 ($query =~ /fast/ && $query =~ /light/ && $query =~ /travel|move/)) {
	insert_answers($Aranea, "internal", ["299,792,458 meters per second"]);
    }

    if ( $query =~ /How far away is the moon/i or
	 $query =~ /distance to the moon/i ) {
	insert_answers($Aranea, "internal", ["240,000 miles"]);
    }

    my $acronyms_signature = sub {
	my $s = shift;

	if ( $s =~ /([A-Za-z.&]+) stand for/i ) {
	    my $acr = uc($1);
	    $acr =~ s/\.//g;
	    return $acr
	}
	return '';
    };

    my $currency_signature = sub {
	my $s = shift;

	my $cpat = '([A-Z][a-z]+( (and|[A-Z][a-z]+))*)';
	return '' unless $s =~ /currency|money/i;
	return '' unless $s =~ /use/i;

	$s =~ s/^[A-Z][a-z]+//g;

	return $& if $s =~ /$cpat/;
	return '';
    };

    if ( my $arg = &$acronyms_signature($query) ) {
	my @results = Aranea::Resources::Acronyms::get($arg);
	insert_answers($Aranea, "Acronyms", \@results);
    }

    if ( my $arg = &$currency_signature($query) ) {
	my @results = Aranea::Resources::Infoplease::get($arg, "currency");
	insert_answers($Aranea, "Infoplease", \@results);
    }

    my $arg;
    db_insert($Aranea, "Infoplease", [$arg, "area"])
	if $arg = _country_area_sig($query);

    db_insert($Aranea, "Infoplease", [$arg, "population"])
	if $arg = _country_population_sig($query);

    db_insert($Aranea, "Infoplease", [$arg, "capital"])
	if $arg = _country_capital_sig($query);

    _answer_50States($Aranea, $query);
    _answer_WebElements($Aranea, $query);

    return $Aranea;
}
#>

#< sub _country_area_sig
sub _country_area_sig {
    my $s = shift;

    my $cpat = '([A-Z][a-z]+( (and|[A-Z][a-z]+))*)';
    return $1 if $s =~ /How big is (?:the )?$cpat/;
    return $1 if $s =~ /size of (?:the )?$cpat\?/;
    return $1 if $s =~ /area of (?:the )?$cpat/;
    return $1 if $s =~ /What is $cpat(?:'s) area/;

    return '';
};
#>
#< sub _country_population_sig
sub _country_population_sig {
    my $s = shift;

    my $cpat = '([A-Z][a-z]+( (and|[A-Z][a-z]+))*)';
    return $1 if $s =~ /population of (?:the )?$cpat/;
    return $1 if $s =~ /$cpat(?:'s) (?:total )?population/;
    return $1 if $s =~ /people live in $cpat/;

    return '';
};
#>
#< sub _country_capital_sig
sub _country_capital_sig {
    my $s = shift;
    my $capital = '';

    my $cpat = '([A-Z][a-z]+( (and|[A-Z][a-z]+))*)';
    $capital = $1 if $s =~ /capital (?:city )?of (?:the )?$cpat/;
    $capital = $1 if $s =~ /$cpat(?:'s)? capital/;

    # handle slightly awkward, but nevertheless plausible questions
    # like 'What is the Haitian capital?'

    $capital = $ADJ_TO_LOCATION{$capital}->[0]
	if $capital and exists $ADJ_TO_LOCATION{$capital};

    return $capital ? $capital : '';
};
#>

#< sub _answer_50States
# Code for answering various questions about states, from
# 50states.com: area, bird, capital, flower, motto, nickname,
# population, tree
sub _answer_50States {
    my ($Aranea, $query) = @_;
    my $arg;

    db_insert($Aranea, "50states", [$arg, "area"])
	if $arg = _state_area_sig($query);

    db_insert($Aranea, "50states", [$arg, "bird"])
	if $arg = _state_bird_sig($query);

    db_insert($Aranea, "50states", [$arg, "capital"])
	if $arg = _state_capital_sig($query);

    db_insert($Aranea, "50states", [$arg, "flower"])
	if $arg = _state_flower_sig($query);

    db_insert($Aranea, "50states", [$arg, "motto"])
	if $arg = _state_motto_sig($query);

    db_insert($Aranea, "50states", [$arg, "nickname"])
	if $arg = _state_nickname_sig($query);

    db_insert($Aranea, "50states", [$arg, "population"])
	if $arg = _state_population_sig($query);

    db_insert($Aranea, "50states", [$arg, "tree"])
	if $arg = _state_tree_sig($query);
}
#>
#< sub _state_tree_sig

sub _state_tree_sig {
    my $s = shift;

    return $1 if $s =~ /($STATE)(?:'s)? state tree/i;
    return $1 if $s =~ /state tree of ($STATE)/i;

    return '';
};

#>
#< sub _state_bird_sig
sub _state_bird_sig {
    my $s = shift;

    return $1 if $s =~ /($STATE)(?:'s)? state bird/i;
    return $1 if $s =~ /state bird of ($STATE)/i;

    return '';
};
#>
#< sub _state_flower_sig
sub _state_flower_sig {
    my $s = shift;

    return $1 if $s =~ /($STATE)(?:'s)? state flower/i;
    return $1 if $s =~ /state flower of ($STATE)/i;

    return '';
};
#>
#< sub _state_nickname_sig
sub _state_nickname_sig {
    my $s = shift;

    return $1 if $s =~ /($STATE)(?:'s)? (?:state )?nickname/i;
    return $1 if $s =~ /nickname of ($STATE)/i;

    return '';
};
#>
#< sub _state_motto_sig
sub _state_motto_sig {
    my $s = shift;

    return $1 if $s =~ /($STATE)(?:'s)? (?:state )?motto/i;
    return $1 if $s =~ /motto (?:of|for) ($STATE)/i;

    return '';
};
#>
#< sub _state_population_sig
sub _state_population_sig {
    my $s = shift;

    return $1 if $s =~ /($STATE)(?:'s)? population/i;
    return $1 if $s =~ /population of ($STATE)/i;
    return $1 if $s =~ /people live in ($STATE)/i;

    return '';
};
#>
#< sub _state_capital_sig
sub _state_capital_sig {
    my $s = shift;

    return $1 if $s =~ /($STATE)(?:'s)? capital/i;
    return $1 if $s =~ /capital (city )?of ($STATE)/i;

    return '';
};
#>
#< sub _state_area_sig
sub _state_area_sig {
    my $s = shift;

    return $1 if $s =~ /($STATE)(?:'s)? area/i;
    return $1 if $s =~ /area of ($STATE)/i;
    return $1 if $s =~ /size of ($STATE)\?/i;
    return $1 if $s =~ /How big is ($STATE)/i;

    return '';
};
#>

#< sub _answer_WebElements
my $ELEMENTS = $Aranea::Resources::WebElements::WEBELEMENTS;

sub _answer_WebElements {
    my ($Aranea, $query) = @_;
    my $arg;
    
    db_insert($Aranea, "WebElements", [$arg, "symbol"])
	if $arg = _elements_symbol_sig($query);
    
    db_insert($Aranea, "WebElements", [$arg, "discoverer"])
	if $arg = _elements_discoverer_sig($query);
    
    db_insert($Aranea, "WebElements", [$arg, "discovery_date"])
	if $arg = _elements_discovery_date_sig($query);

    db_insert($Aranea, "WebElements", [$arg, "atomic_number"])
	if $arg = _elements_atomic_number_sig($query);

    db_insert($Aranea, "WebElements", [$arg, "atomic_mass"])
	if $arg = _elements_atomic_weight_sig($query);

    db_insert($Aranea, "WebElements", [$arg, "melting_point"])
	if $arg = _elements_melting_point_sig($query);

    db_insert($Aranea, "WebElements", [$arg, "boiling_point"])
	if $arg = _elements_boiling_point_sig($query);
}
#>
#< sub _elements_symbol_sig
sub _elements_symbol_sig {
    my $s = shift;

    return $2 
	if $s =~ /^What is/i and $s =~ /(chemical )?symbol for $ELEMENTS/i;

    return $1
	if $s =~ /^What is/i and $s =~ /$ELEMENTS\'s (chemical )?symbol/i;

    return '';
};
#>
#< sub _elements_discoverer_sig
sub _elements_discoverer_sig {
    my $s = shift;

    return $1
	if $s =~ /(^Who|person)/i and $s =~ /discovered $ELEMENTS( |\W|$)/i;

    return '';
};
#>
#< sub _elements_discovery_date_sig
sub _elements_discovery_date_sig {
    my $s = shift;

    return $2 if $s =~ /(^When|date|year)/i and 
      	         $s =~ /( |\W|^)$ELEMENTS discovered/i;

    return '';
};
#>
#< sub _elements_atomic_number_sig
sub _elements_atomic_number_sig {
    my $s = shift;
    
    return $1 if $s =~ /^What is/i and 
	         $s =~ /atomic number of $ELEMENTS( |\W|$)/i;
    
    return $2 if $s =~ /^What is/i and 
	         $s =~ /( |\W|^)$ELEMENTS\'s atomic number/i;

    return '';
};
#>
#< sub _elements_atomic_weight_sig
sub _elements_atomic_weight_sig {
    my $s = shift;

    return $2 if $s =~ /^What is/i and 
                 $s =~ /atomic (weight|mass) of $ELEMENTS( |\W|$)/i;

    return $2 if $s =~ /^What is/i and 
                 $s =~ /( |\W|^)$ELEMENTS\'s atomic (weight|mass)/i;

    return '';
};
#>
#< sub _elements_melting_point_sig
sub _elements_melting_point_sig {
    my $s = shift;

    return $1 if $s =~ /(^What is|temperature|^How hot)/i and
	         $s =~ /(melting point|melt)/i and
	         $s =~ / $ELEMENTS( |\W|$)/i;

    return '';
};
#>
#< sub _elements_boiling_point_sig
sub _elements_boiling_point_sig {
    my $s = shift;

    return $1 if $s =~ /(^What is|temperature|^How hot)/i and
	         $s =~ /(boiling point|boil)/i and
		 $s =~ / $ELEMENTS( |\W|$)/i;

    return '';
};
#>

#< sub insert_answers
sub insert_answers {
    my ($Aranea, $source, $results) = @_;

    foreach my $ans ( @$results ) {
	next unless $ans;  # don't insert null answers
	my $entry = Aranea::XML::CandidateEntry->new($ans, 999999);
	$entry->add_support($source);
	$Aranea->candidate_list()->add_entry($entry);
    }
}
#>

1;
