#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/ProcessQuestion.pm,v 1.11 2005/06/10 16:17:16 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::Modules::ProcessQuestion>> - Aranea module for
preprocessing the question

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

This Aranea module preprocess the question by expanding contractions,
finding the focus, and POS tags it.

=cut

###############################################################################
#>
#< Header
package Aranea::Modules::ProcessQuestion;

use strict;
use warnings;
use lib qw(lib pm);                         # needed for command-line mode

use IO::Handle;

use Aranea::XML::Aranea;
use Aranea::Common qw(:default);

#>

#< STDIN handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    $Aranea->insert_pipeline_tag('$Id: ProcessQuestion.pm,v 1.11 2005/06/10 16:17:16 jimmylin Exp $ ');

    # save the original query and id
    my $q = $Aranea->query()->original();

    $q =~ s/^(Approximately|About|Around) //i;

    # expand contractions
    $q =~ s/^What's/What is/i;
    $q =~ s/^Who's/Who is/i;
    $q =~ s/^Where's/Where is/i;
    $q =~ s/^When's/When is/i;

    # Don't do this: sometimes it actually hurts, e.g., q444 "Where is
    # the location of the Orange Bowl?"
    #$q =~ s/^Where is the location of/Where is/i;

    $q =~ s/^In what place/Where/i;
    $q =~ s/^On what date/When/i;

    # q245 "Where can you find the Venus flytrap?"
    if ( $q =~ /^Where can you find/i ) {
	$q =~ s/^Where can you find/Where can/i;
	$q =~ s/\W+$//g;
	$q .= " be found?";
    }

    # q1244 What is the date of Mexico's independence?
    $q =~ s/^What is the date of/When is/i;

    $q =~ s/^What date/When/i;
    $q =~ s/^What day/When/i;
    $q =~ s/^Tell me wh/Wh/i;

    # 'Name a X' questions should have the imperative dropped.
    $q =~ s/^Name //i;

    # U.S.A. leads to awkward questions
    #   e.g., q1304 Which U.S.A. president appeared on "Laugh-In"?
    $q =~ s/U\.?S\.?A\.?/U.S./;
    # NOTE: don't make the match case insensitive!

    # expands 'U.S.' when used as a PP object
    #   e.g., q275 Who was the first vice president of the U.S.?
    $q =~ s/of the U\.S\./of the United States/i;

    # the following tweak was verified to help somewhat, but the
    # sample size was too small too determine whether this effect was
    # statistically significant.
    $q =~ s/the names? of //i;

    # POS tag the query
    (my $tq = $q) =~ s/\W$//g;    # trim trailing punctuation
    my $tagged = pos_tag($tq);

    if ( $tagged =~ /^([^\/]+)\/IN wh/i ) {
	my $prep = lc($1);
	$tagged =~ s/^$1\/[^ ]+ wh/Wh/;
	$tagged .= "$prep/IN";

	$q =~ s/^$prep wh/Wh/i;
	$q =~ s/(\W+)?$/ $prep$1/;
	$q =~ s/ +/ /g;
    }
    $Aranea->query()->tagged($tagged);

    # the modified query
    $Aranea->query()->text($q);

    # extract the question focus, if possible
    if ( $tagged =~ /^How\/[^ ]+ many\/[^ ]+ /i ) {
	my $qfocus = $';

	if ( $qfocus =~ /^([^\/]+\/NN[A-Z]* )+/ ) {
	    $qfocus = strip_pos_tags($&);
	    $Aranea->query()->qfocus($qfocus);
	}
    }

    return $Aranea;
}
#>

1;
