#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/ScanResults.pm,v 1.17 2005/06/10 16:17:16 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::Modules::ScanResults>> - Aranea module for generating
initial answer candidates from raw snippets.

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

This Aranea module postprocesses snippets stored in the C<page_list>,
completing the requests in the C<request_list>.  Answer candidates are
generated and stored in the C<candidate_list>.  With an inexact
request, the snippets are broken down by "..." (which indicates
distinct passages), and converted into candidates.  With an exact
request, the text fragment satisfying the imposed constraints is
instantiated as the answer candidate.

=cut

###############################################################################
#>
#< Header
package Aranea::Modules::ScanResults;

use strict;
use warnings;
use lib qw(lib pm);                         # needed for command-line mode

use IO::Handle;

use Aranea::XML::Aranea;

#>

#< STDIN handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    $Aranea->insert_pipeline_tag('$Id: ScanResults.pm,v 1.17 2005/06/10 16:17:16 jimmylin Exp $ ');

    # if there are no requests, then there is nothing to do
    return $Aranea if $Aranea->request_list()->entries_count() == -1;

    my $request_entries = $Aranea->request_list()->entries();

    my $i=0;
    for my $request ( @$request_entries ) {
	_process_request($Aranea, $request, $i);

	$i++;
    }

    return $Aranea;
}
#>

#< sub _process_request
sub _process_request {
  my ($Aranea, $request, $id) = @_;

    my $entries = $Aranea->page_list()->entries();

    my $score = $request->score();
    my $query = $request->query();
    my $max_length = $request->get_constraint('max_length');
    my $max_words = $request->get_constraint('max_words');
    my $use_backoff = $request->get_constraint('use_backoff');

    # convert the query into a regular expression
    my $regex = _convert_query_to_regexp($query);

    # hash for keeping track of unique page titles.  We want to add
    # the page titles as candidates, but only once per title.
    my %titlehash = ();

    # When truncating with max_words or max_length, decide whether to
    # truncate from the beginning or end of the string.
    my $keep_beginning = $query =~ m/\?[a-zA-Z]+[.]?\z/;

    for my $entry ( @$entries ) {
	# make sure that the snippet id matches the request id
	next if $entry->id() != $id;

	my $candidate = '';
	if ( $use_backoff ) {
	    # If we are using backoff, then use the entire page summary as
	    # the candidate.  However, split the summaries by "...", which
	    # marks disjoint fragments.

	    $candidate = $entry->page_summary();

	    my @fragments = split(/\.\.\./, $candidate);

	    foreach my $f ( @fragments ) {
		$f =~ s/^ +//g;
		$f =~ s/ +$//g;
		next if $f eq '';

		my $candidate_list_entry =
		  Aranea::XML::CandidateEntry->new($f, $score);
		$Aranea->candidate_list()->add_entry($candidate_list_entry);
	    }

	    my $title = $entry->page_title();
	    unless ( exists($titlehash{$title}) ) {
		my $title_entry =
		    Aranea::XML::CandidateEntry->new($title, $score);
		$Aranea->candidate_list()->add_entry($title_entry);
		$titlehash{$title} = 1;
	    }
	}
	else {
	    # We are not using backoff, so we have to find the extract
	    # fragment that binds to ?x

	    # first get a rough match
	    $candidate = _get_match($regex, $entry->page_summary());

	    # satisfy the max_word constraint
	    my @words = split(/\s+/, $candidate);
	    if ( scalar(@words) > $max_words ) {
		# keep trimming of words (off back or front) until
		# constraint satisftied

		if ( $keep_beginning ) {
		    $candidate = join(' ', splice(@words, 0, $max_words));
		}
		else {
  		    $candidate = join(' ', splice(@words,
						  scalar(@words) - $max_words,
						  scalar(@words)));
		}
	    }

	    # satisfy the max_length constraint
	    while ( length($candidate) > $max_length ) {
		# chop off either leading or trailing tokens one by
		# one util the length constraint is satisfied.  (break
		# candidates up only at word boundaries)

		my @words = split(/\s+/,$candidate);

		last if scalar( @words ) == 0;

		# chop either leading or trailing token
		$keep_beginning ? pop(@words) : shift(@words);
		$candidate = join(' ', @words);
	    }

	    # kill leading and trailing spaces
	    $candidate =~ s/^\s+//g;
	    $candidate =~ s/\s+$//g;

	    # finally, add the candidate
	    if ( $candidate ) {
		my $candidate_list_entry =
		    Aranea::XML::CandidateEntry->new($candidate, $score);
		$Aranea->candidate_list()->add_entry($candidate_list_entry);
	    }
	}
    }
}
#>
#< sub _get_match
sub _get_match {
    my ($regex, $text) = @_;

    if ( $text =~ m/$regex/i ) {
	return $1;
    }

    return 0;
}
#>
#< sub _convert_query_to_regexp
# convert the query into a regular expression
sub _convert_query_to_regexp {
    my $query = shift;
    $query =~ s/([.])/[.]/g;

    my $whiteSpaceRegex = "\\s*";

    # replace spaces with the white space
    # regular expression
    $query =~ s/(\s+)/$whiteSpaceRegex/g;

    #get rid of extra whitespace
    $query =~ s/\s//g;

    # convert "?x" into a regular expression
    $query =~ s/\?[a-zA-Z]+/([a-zA-Z 0-9.,']*)/g;  # ']);

    return $query;
}
#>

1;
