#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/SearchGoogleAPI.pm,v 1.9 2005/06/10 16:17:17 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::Modules::SearchGoogleAPI>> - Aranea module for fetching
relevant pages from Google

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

The difference between this module and
C<Aranea::Modules::SearchGoogle> is that this module utilizes the
Google API.

=cut

###############################################################################
#>
#< Header
package Aranea::Modules::SearchGoogleAPI;

use strict;
use warnings;
use lib qw(lib pm);                         # needed for command-line mode

use HTML::Entities;
use URI::URL;
use IO::Handle;

use Aranea::XML::Aranea;
use Aranea::XML::PageEntry;
use Aranea::Common qw(:default);

# The following snippet of code was taken from adapted from:
#
# GoogleSearch.pl
# 11apr2002 - matt@interconnected.org http://interconnected.org/home/
# Demonstrates use of the doGoogleSearch method on the Google API.
# See http://www.google.com/apis/ to get your key and download the WSDL
#(which this script expects to find in its directory).

use SOAP::Lite;

# Configuration
our $key = "INSERT_KEY_HERE";

# Redefine how the default deserializer handles booleans.
# Workaround because the 1999 schema implementation incorrectly doesn't
# accept "true" and "false" for boolean values.
# See http://groups.yahoo.com/group/soaplite/message/895
*SOAP::XMLSchema1999::Deserializer::as_boolean =
    *SOAP::XMLSchemaSOAP1_1::Deserializer::as_boolean = 
    \&SOAP::XMLSchema2001::Deserializer::as_boolean;

our $wsdl = find_file("Aranea/Modules/GoogleSearch.wsdl");

# Initialise with local SOAP::Lite file
our $service = SOAP::Lite->service("file:$wsdl");

# End of adapted code snippet

#>

#< STDIN handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    $Aranea->insert_pipeline_tag('$Id: SearchGoogleAPI.pm,v 1.9 2005/06/10 16:17:17 jimmylin Exp $ ');

    # if there are no requests, then there is nothing to do
    return $Aranea if $Aranea->request_list()->entries_count() == -1;

    my $request_entries = $Aranea->request_list()->entries();

    my $i=0;
    for my $request ( @$request_entries ) {
	_parse_query($Aranea, $request, $i);
	$i++;
    }

    return $Aranea;
}
#>

#< sub _parse_query
sub _parse_query {
    my ($Aranea, $request, $id) = @_;

    my $pages_to_search = $request->get_constraint("pages_to_search");

    my $pages_per_chunk = 10;
    my $total_pages = 0;
    my $start_page = 0;
    my $prev_total_pages = 0;

    my $debug_flag = 0;

    while ( $total_pages < $pages_to_search ) {
	print STDERR "Starting at entry $start_page, " .
	    " attempting to grab $pages_per_chunk\n"
		if $debug_flag;

	my $entries_retrieved =
	    _get_google_pages($pages_per_chunk, $start_page, $request,
			      $Aranea, $id);

	$total_pages += $entries_retrieved;

	if ( $debug_flag ) {
	    print STDERR "Sucessfully grabbed $entries_retrieved.\n";
	    print STDERR "In total, $total_pages have been retrieved.\n\n";
	}

	# if the Google API returned less than 10 entries, then it
	# means were out of hits
	last if $entries_retrieved < 10;

	$start_page = $total_pages + 1;
	$prev_total_pages = $total_pages;
    }
}
#>
#< sub _get_google_pages
sub _get_google_pages {
    my ($pages_to_search, $start_page, $request, $Aranea, $id) = @_;

    my %sighash = ();    # hash for unique "page signatures"
    my %sigurl = ();     # hash for unique URLs (see comments below)

    my $result;
    eval {
	$result = $service
	    ->doGoogleSearch($key,                      # key
			     _prepare_query($request),  # search query
			     $start_page,               # start results
			     10,                        # max results
			     "true",                    # filter: boolean
			     "",                        # restrict (string)
			     "false",                   # safeSearch: boolean
			     "lang_en",                 # lr
			     "latin1",                  # ie
			     "latin1"                   # oe
			    );
    };

    if ( $! ) {
	eval {
	    $result = $service->doGoogleSearch($key, _prepare_query($request), $start_page, 10, "true", "", "false", "lang_en", "latin1", "latin1");
	};
    }

    return 0 unless defined($result->{resultElements});

    my $page_list = $Aranea->page_list();
    my $total_entries = 0;

    for my $i ( 0 .. $#{@{$result->{resultElements}}} ) {
	my $page_title = _clean($result->{resultElements}[$i]->{title});
	my $page_url = _clean($result->{resultElements}[$i]->{URL});
	my $page_summary = _clean($result->{resultElements}[$i]->{snippet});

	#print "PAGE_TITLE = ##$page_title##\n";
	#print "PAGE_URL = ##$page_url##\n";
	#print "PAGE_SUMMARY = ##$page_summary##\n";
	#print "\n\n";

	my $entry = Aranea::XML::PageEntry->new();
	$entry->page_url($page_url);
	$entry->page_title($page_title);
	$entry->page_summary($page_summary);
	$entry->id($id);

	# This below code tries to avoid grabbing duplicate pages at
	# different URLs.  This is done by computing a "page
	# signature" and throwing out pages with the same signature.
	# For now, the signature is the page summary with all
	# non-alphanumeric characters removed.  Obviously, a smarter
	# page signature would be nice...

	# June 21, 2002 -- very interesting point: if multiple
	# requests retrieve the same page, should we have duplicate
	# entries?  without rigourous testing, the answer seems to be
	# yes... although there are sometimes screw cases.

	# August 1, 2002 -- actually, the answer appears to be no!
	# (not from rigorous regression testing, but from ad-hoc
	# experimentation) if multiple requests get the same page,
	# then we should keep those duplicate pages.

	# As written, the below code for $sigurl is uncessary, since
	# Google is supposed to filter duplicate pages. (note that we
	# declare $sigurl as local to this procedure.) The device is
	# for ensuring unique urls across multiple requests.

	my $page_sig = $page_summary;
	$page_sig =~ s/[^A-Za-z0-9]//g;

	if ( not exists($sighash{$page_sig}) and
	     not exists($sigurl{$page_url}) ) {

	    # we don't want pages that have TREC questions as samples
	    # on them... naturally, this won't be an issue with fresh
	    # questions in future competitions.

	    if ( $page_url !~ /TREC/i and $page_summary !~ /TREC/i ) {

		#push(@$entries, $entry);
		$page_list->add_entry($entry);
		$total_entries++;

		$sighash{$page_sig} = 1;
		$sigurl{$page_url} = 1;
	    }
	}
    }

    return $total_entries;
}
#>
#< sub _clean
sub _clean {
    my $s = shift;

    $s =~ s/[\r\n]+//g;
    $s =~ s/<[^>]+>/ /g;
    $s =~ s/\s+/ /g;

    return $s;
}
#>
#< sub _prepare_query
sub _prepare_query {
    my $request = shift;

    my $query = $request->query();
    my $use_backoff = $request->get_constraint("use_backoff");

    # Parse the search into regular expression so google query can use
    # it.  Right now, it gets rid of "?" terms and put + in front of
    # every word (if not using backoff).  this will insulate us from
    # google changing the list of their common words.

    $query =~ s/[ \n]+$//g;
    $query =~ s/\?[a-zA-Z]+//g;
    $query =~s/^\s+//;             # get rid of spaces in front
    $query =~s/\s+$//;             # get rid of trailing white spaces

    # if we're not using backoff, add + in front of every word
    unless ( $use_backoff ) {
	$query=~s/^/\+/;
	$query =~s/\s/ \+/g;
    }

    $query = "\"$query\"" unless $use_backoff;

    return $query;
}
#>

1;
