#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/Modules/SearchTeoma.pm,v 1.6 2005/06/11 17:41:52 jimmylin Exp $
#
#>
#< POD

###############################################################################

=head1 NAME

B<C<Aranea::Modules::SearchTeoma>> - Aranea module for fetching
relevant pages from Teoma

=head1 SYNOPSIS

This Aranea modules is meant to be used in conjunction with the
L<C<Aranea::Shell>|Aranea::Shell>.

=head1 DESCRIPTION

This Aranea module partially executes the requests in the
C<request_list> by fetching raw text snippets from Teoma and
populating the C<page_list>.  Note that this module B<does not>
performs additional postprocessing, e.g., to extract bindings for
exact requests.  For further postprocessing of raw Teoma extracts,
see L<C<ScanResults>|Aranea::Modules::ScanResults>.  The idea is to
decouple the snippet extraction code from the candidate generation
code so that we can replace Teoma with any other search engine.

=cut

###############################################################################

#>
#< Header
package Aranea::Modules::SearchTeoma;

use strict;
use warnings;
use lib qw(lib pm);                         # needed for command-line mode

use HTML::Entities;
use URI::URL;
use IO::Handle;

use Aranea::XML::Aranea;
use Aranea::XML::PageEntry;
use Aranea::Common qw(:default);

#>

#< STDIN handler
unless ( caller() ) {
    my $stdin = new_from_fd IO::Handle(fileno(STDIN),"r");
    my $Aranea = Aranea::XML::Aranea->new($stdin);

    $Aranea = doModule($Aranea);

    $Aranea->write_xml();
}
#>

#< sub doModule
sub doModule {
    my $Aranea = shift;

    $Aranea->insert_pipeline_tag('$Id: SearchTeoma.pm,v 1.6 2005/06/11 17:41:52 jimmylin Exp $ ');

    # if there are no requests, then there is nothing to do
    return $Aranea if $Aranea->request_list()->entries_count() == -1;

    my $request_entries = $Aranea->request_list()->entries();

    my $i=0;
    for my $request ( @$request_entries ) {
	_parse_query($Aranea, $request, $i);
	$i++;
    }

    return $Aranea;
}
#>

#< sub _parse_query
sub _parse_query {
    my ($Aranea, $request, $id) = @_;

    my $pages_to_search = $request->get_constraint("pages_to_search");

    my $pages_per_chunk = 100;
    my $total_pages = 0;
    my $start_page = 0;
    my $prev_total_pages = 0;

    my $debug_flag = 0;

    while ( $total_pages < $pages_to_search ) {

      # Jan. 15, 2002 -- I decided to always try and grab 100 entries
      # at a time...
      #if ( $total_pages + $pages_per_chunk > $pages_to_search ) {
	#$pages_per_chunk = $pages_to_search - $total_pages;
      #}

      print STDERR "Starting at entry $start_page, " .
	" attempting to grab $pages_per_chunk\n"
	  if $debug_flag;

      my ($num_results, $entries_retrieved) =
	_get_teoma_pages($pages_per_chunk, $start_page, $request,
			  $Aranea, $id);

      $total_pages += $entries_retrieved;

      if ( $debug_flag ) {
	print STDERR "Teoma reports $num_results results to this query\n";
	print STDERR "Sucessfully grabbed $entries_retrieved.\n";
	print STDERR "In total, $total_pages have been retrieved.\n\n";
      }

      # This is a general purpose escape hatch: if we aren't getting
      # any new pages (for whatever reason), then give up.  Also, when
      # you try to get a lot of snippets (e.g., 1k), sometimes Teoma
      # just doesn't give you enough results... so give up if we
      # aren't getting more than 10 pages per query.
      last unless ( $prev_total_pages <= $total_pages - 10 );
      last if ( $total_pages == 0 or $pages_to_search > $num_results );

      $start_page = $total_pages + 1;
      $prev_total_pages = $total_pages;
    }
}
#>
#< sub _get_teoma_pages

sub _get_teoma_pages {
    my ($pages_to_search, $start_page, $request, $Aranea, $id) = @_;

    my %sighash = ();    # hash for unique "page signatures"
    my %sigurl = ();     # hash for unique URLs (see comments below)

    my $uri = _prepare_url($request, $pages_to_search, $start_page);
    #print "uri:$uri<br>";
    #Aranea::Common::set_page_delay(1);
    my $content = safe_get($uri);
    $content = demoronize($content);

    my ($num_results) = $content =~ m/of about ([,0-9]+?):/s;
    $content = $';
    $num_results =~ s/,//g if $num_results;
    $num_results = 0 unless $num_results;

    return ($num_results, 0) unless ($num_results > 0);

    my $page_list = $Aranea->page_list();
    my $total_entries = 0;

    my @raw_snippets = split /<div\s+?id=\"result\"\s*?>/, $content;

    foreach my $c ( @raw_snippets ) {
      my ($page_url) = $c =~ m|<a .*?u=(.*?)".*?>|s;
      next unless $page_url;
      $c = $';

      # find the title of the page
      my ($page_title) = $c =~ m|(.*?)</a>|s;
      $c = $';

      $page_title =~ s|<b>||g;
      $page_title =~ s|<\/b>||g;
      $page_title =~ s|\n| |g;

      my ($page_summary) = $c =~ m|<div>(.*?)</div>|s;
      next unless $page_summary;
      $c = $';

      # kill all HTML tags
      $page_summary =~ s|<[^>]+>||g;

      # decode entities like &nbsp; &quot; etc.
      $page_summary = decode_entities($page_summary);

      # sometimes, there is no space after punctation
      $page_summary =~ s/([.,])([A-Z])/$1 $2/g;
      $page_summary =~ s|\n| |g;                   # compress newlines
      $page_summary =~ s|  +| |g;  	           # compress multiple spaces

      # get the page category (cross-indexed from Yahoo), if available
      my $page_category = '';

      #print "PAGE_URL = ##$page_url##\n<br>";
      #print "PAGE_TITLE = ##$page_title##\n<br>";
      #print "PAGE_SUMMARY = ##$page_summary##\n<br>";
      #print "PAGE_CATEGORY = ##$page_category##\n<br><br>";

      my $entry = Aranea::XML::PageEntry->new();
      $entry->page_url($page_url);
      $entry->page_title($page_title);
      $entry->page_summary($page_summary);
      $entry->page_category($page_category) if $page_category;
      $entry->id($id);

      # This below code tries to avoid grabbing duplicate pages at
      # different URLs.  This is done by computing a "page
      # signature" and throwing out pages with the same signature.
      # For now, the signature is the page summary with all
      # non-alphanumeric characters removed.  Obviously, a smarter
      # page signature would be nice...

      # June 21, 2002 -- very interesting point: if multiple
      # requests retrieve the same page, should we have duplicate
      # entries?  without rigourous testing, the answer seems to be
      # yes... although there are sometimes screw cases.

      # August 1, 2002 -- actually, the answer appears to be no!
      # (not from rigorous regression testing, but from ad-hoc
      # experimentation) if multiple requests get the same page,
      # then we should keep those duplicate pages.

      # As written, the below code for $sigurl is uncessary, since
      # Teoma is supposed to filter duplicate pages. (note that we
      # declare $sigurl as local to this procedure.) The device is
      # for ensuring unique urls across multiple requests.

      my $page_sig = $page_summary;
     # print "summary: $page_sig";
      $page_sig =~ s/[^A-Za-z0-9]//g;

      if ( not exists($sighash{$page_sig}) and
	   not exists($sigurl{$page_url}) ) {
     
	# we don't want pages that have TREC questions as samples
	# on them... naturally, this won't be an issue with fresh
	# questions in future competitions.

	if ( $page_url !~ /TREC/i and $page_summary !~ /TREC/i ) {

	  #push(@$entries, $entry);
	  $page_list->add_entry($entry);
	  $total_entries++;

	  $sighash{$page_sig} = 1;
	  $sigurl{$page_url} = 1;
	}
      }
    }

    #print "num: $num_results total:$total_entries";
    return ($num_results, $total_entries);
}

#>
#< sub _prepare_url
sub _prepare_url {
    my ($request, $pages_to_search, $start_page) = @_;

    my $query = $request->query();
    my $use_backoff = $request->get_constraint("use_backoff");

    # Parse the search into regular expression so teoma query can use
    # it.  Right now, it gets rid of "?" terms and put + in front of
    # every word (if not using backoff).  this will insulate us from
    # teoma changing the list of their common words.

    $query =~ s/[ \n]+$//g;
    $query =~ s/\?[a-zA-Z]+//g;
    $query =~s/^\s+//;             # get rid of spaces in front
    $query =~s/\s+$//;             # get rid of trailing white spaces

    ##### if we're not using backoff, add + in front of every word
    unless ( $use_backoff ) {
	$query=~s/^/\+/;
	$query =~s/\s/ \+/g;
    }

    # make a url to teoma's search page
    my $uri = URI::URL->new('http://s.teoma.com/search');

    $query = "\"$query\"" unless $use_backoff;

    $uri->query_form("q" => "$query",
		     "u"=>"$pages_to_search",
		     "page"=>"$start_page",
		     );

    return $uri;
}
#>

1;





