#!/usr/bin/perl -w
#< Boilerplate
#
# The Aranea question answering system
# Copyright (C) 2002-2005, Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, please visit
# http://www.gnu.org/copyleft/gpl.html
#
# $Header: /fs/clip-qa/.cvsroot//Aranea/pm/Aranea/XML/CandidateList.pm,v 1.25 2005/06/08 01:41:47 jimmylin Exp $
#
#>
#< POD
###############################################################################

=head1 NAME

B<C<Aranea::XML::CandidateList>> - object representing the
B<C<E<lt>candidate_listE<gt>>> of an Aranea XML object

=head1 SYNOPSIS

  use Aranea::XML::Aranea;
  use Aranea::XML::CandidateList;
 
  my $stdin = new_from_fd IO::Handle(STDIN, "r");
  my $Aranea = Aranea::XML::Aranea->new($stdin);
 
  my $candidate_list = $Aranea->candidate_list();
  my $entries = $candidate_list->entries();
  # arrayref to entries in candidate list

=head1 DESCRIPTION

The C<CandidateList> object represents the C<candidate_list> section
of an Aranea XML object.  Here is an example:

  <candidate_list>
    <entry>
      <support>
        <doc> http://www.foo.com/p1.html </doc>
        <doc> http://www.bar.com/p1.html </doc>
        <doc> http://www.baz.com/p2.html </doc>
      </support>
      <score>234.287927</score>
      <candidate>foo bar baz</candidate>
    </entry>
    ...
  </candidate_list>

The C<CandidateList> object serves as a container for a collection of
L<C<CandidateEntry>s|Aranea::XML::CandidateEntry>.  Under normal
circumstances, the creation of a C<CandidateList> should be handled
automatically by the L<C<Aranea::XML::Aranea>|Aranea::XML::Aranea>
object.

=head1 METHODS

=over 4

=cut

###############################################################################
#>
#< Header
package Aranea::XML::CandidateList;

use strict;
use warnings;
use Data::Dumper;

use XML::Simple;
use XML::Writer;

use Aranea::XML::CandidateEntry;

#>

#< sub new

=item B<C<new($data)>>

Constructs a C<CandidateList> object given parsed XML data.  Note that
this method is rarely invoked manually; under typical usage patterns,
the constructor of L<C<Aranea::XML::Aranea>|Aranea::XML::Aranea> will
call this method automatically.

=cut
sub new {
    my $proto = shift;
    my $data = shift;
    my $class = ref($proto) || $proto;
    my $self = {};

    # gotta bless it before we can treat it as an object and call
    # methods on it

    bless ($self, $class);

    $self->{entries} = [];

    # interate through the parsed XML data and create each individual
    # CandidateEntry

    foreach my $entry ( @{$data->{entry}} ) {
	my $entryobj = Aranea::XML::CandidateEntry->new($entry);

	# use its own methods to add entries
	$self->add_entry($entryobj);
    }

    return $self;
}
#>

#< sub entries

=item B<C<entries([$listref])>>

Accesses or mutates the collection of
L<C<CandidateEntry>s|Aranea::XML::CandidateEntry>.  When called with
no arguments, this method returns a listref of C<CandidateEntry>s.
When called with a single listref, the collection of
C<CandidateEntry>s is set to the listref.

The typical usage pattern is to iterate over each page object and perform
some action, e.g.,

  my $candidate_list = $Aranea->candidate_list();
  my $entries = $candidate_list->entries();
 
  for my $entry ( @$entries ) {
      print $entry->candidate() . "\n";
 
      # do more stuff...
  }

=cut
sub entries {
    my $self = shift;

    @_ ? $self->{entries} = shift : $self->{entries};
}
#>
#< sub entries_count

=item B<C<entries_count()>>

Returns the number of C<CandidateEntry>s in this
C<CandidateList>.  Note that the return value is similiar to that
of C<$#>, i.e., -1 for no entries, 0 for one entry, etc.

=cut
sub entries_count {
    my $self = shift;

    return $#{@{$self->{entries}}};
}
#>
#< sub add_entry

=item B<C<add_entry($entry)>>

Adds a C<CandidateEntry> to this C<CandidateList>.

=cut
sub add_entry {
    my $self = shift;
    my $entry = shift;

    push @{$self->entries()}, $entry;
}

#>
#< sub map_candidates

=item B<C<map_candidates($proc)>>

Maps a procedure over the answer candidates.  Thie method takes as its
single argument a procedure that take a candidate string as the
argument; the answer candidate is mutated to the results of this
procedure.

  # expand 'Jan' to 'January'
  sub foo { my $s = shift; $s =~ s/Jan/January/g; $s }
  $Aranea->candidate_list()->map_candidates(\&foo);

=cut
sub map_candidates {
    my $self = shift;
    my $proc = shift;
    my $entries = $self->entries();

    # map over candidates
    map { $_->candidate(&{$proc}($_->candidate())) } @$entries;
}
#>
#< sub filter_candidates

=item B<C<filter_candidates($proc)>>

Filter answer candidates.  Thie method takes as its single argument a
procedure that take a candidate string as the argument and evaluates
to true iff the candidate is to be kept (otherwise candidate is
discarded).

  # throw away candidates that do not have 'foo' in them.
  sub foo { shift =~ /foo/ }
  $Aranea->candidate_list()->filter_candidates(\&foo);

=cut
sub filter_candidates {
    my $self = shift;
    my $criteria = shift;
    my $entries = $self->entries();

    # apply the criteria to select the entrie
    my @new_entries = grep { &{$criteria}($_->candidate()) } @$entries;

    # discard the old entries for the new ones
    $self->entries(\@new_entries);
}
#>
#< sub filter_entries

=item B<C<filter_entries($proc)>>

Filter answer entries.  Thie method takes as its single argument a
procedure that take a C<CandidateEntry> as an argument and evaluates
to true iff the candidate is to be kept (otherwise candidate is
discarded).  The major difference between this method and
C<filter_candidates> is that the criteria procedure in
C<filter_candidates> is passed only the candidate string, whereas
C<filter_entries> allows the criteria procedure access to the entire
C<CandidateEntry>, e.g., the score and supporting documents.

=cut
sub filter_entries {
    my $self = shift;
    my $criteria = shift;
    my $entries = $self->entries();

    # apply the criteria to select the entries
    my @new_entries = grep { &{$criteria}($_) } @$entries;

    # discard the old entries for the new one
    $self->entries(\@new_entries);
}
#>
#< sub write_xml

=item B<C<write_xml($writer)>>

Writes the XML representation of this C<CandidateList>; takes an
C<XML::Writer> as its only argument.  Note that this method is rarely
invoked manually; under typical usage patterns, the C<write_xml>
method of L<C<Aranea::XML::Aranea>|Aranea::XML::Aranea> will call this
method automatically.

=cut
sub write_xml {
    my $self = shift;
    my $writer = shift;

    # do nothing if there aren't any entries
    return if $self->entries_count() == -1;

    # get and sort all entries
    my $entries = $self->entries();
    @$entries = sort {$a->{score} <=> $b->{score}} @$entries;

    #print STDERR Dumper($entries);
    #print STDERR XMLout($entries, 'noattr' => 1);

    $writer->startTag('candidate_list');
    for my $entry ( @$entries ) {           # iterate through all entries...
	$writer->startTag('entry');
	$entry->write_xml($writer);         # dispatch on CandidateEntry method
	$writer->endTag('entry');

    }
    $writer->endTag('candidate_list');
}
#>

1;
